@@ -6,7 +6,6 @@ import select
 import struct
 import socket
 import Queue
-import copy
 import time
 import os
 import io
@@ -21,34 +20,32 @@ from pyroute2.netlink.generic import NETLINK_GENERIC
 from pyroute2.netlink.generic import NETLINK_UNUSED
 
 
+_QUEUE_MAXSIZE = 4096
+
+
 class NetlinkError(Exception):
     '''
     Base netlink error
     '''
-    # FIXME move to a dedicated module
-    pass
-
-
-class NetlinkSocketError(socket.error):
-    '''
-    Socket error
-    '''
     def __init__(self, code, msg=None):
         msg = msg or os.strerror(code)
-        super(NetlinkSocketError, self).__init__(code, msg)
+        super(NetlinkError, self).__init__(code, msg)
         self.code = code
 
 
-class NetlinkQueueEmpty(NetlinkError):
-    '''
-    Timeout reached on a message queue polling
-    '''
-    def __init__(self, key):
-        self.key = key
+##
+# FIXME: achtung, monkeypatch!
+#
+# Android QPython 2.7 platform has no AF_UNIX, so add some
+# invalid value just not to fail on checks.
+#
+if not hasattr(socket, 'AF_UNIX'):
+    socket.AF_UNIX = 65535
 
 
 def _monkey_handshake(self):
     ##
+    # FIXME: achtung, monkeypatch!
     #
     # We have to close incoming connection on handshake error.
     # But if the handshake method is called from the SSLSocket
@@ -77,6 +74,8 @@ def _monkey_handshake(self):
 ssl.SSLSocket.do_handshake = _monkey_handshake
 
 
+AF_PIPE = 255  # Right now AF_MAX == 40
+
 ## Netlink message flags values (nlmsghdr.flags)
 #
 NLM_F_REQUEST = 1    # It is request message.
@@ -110,10 +109,26 @@ IPRCMD_RELOAD = 5
 IPRCMD_ROUTE = 6
 
 
-def _get_socket(url, server_side, ssl_keys=None):
-    assert type(url) is str
+def _get_plugin(url):
+    scheme, command = url.split('://')
+    name = 'plugin_%s' % (scheme)
+    plugins = __import__('pyroute2.netlink.plugins',
+                         globals(),
+                         locals(),
+                         [name],
+                         -1)
+    plugin = getattr(plugins, name)
+    return (plugin.plugin_init['type'],
+            plugin.plugin_init['create'],
+            command)
+
+
+def _get_socket(url, server_side=False, ssl_keys=None):
+    assert url[:6] in ('tcp://', 'ssl://', 'tls://') or \
+        url[:11] in ('unix+ssl://', 'unix+tls://') or url[:7] == 'unix://'
     target = urlparse.urlparse(url)
     hostname = target.hostname or ''
+    ssl_keys = ssl_keys or {}
     use_ssl = False
     ssl_version = 2
 
@@ -156,12 +171,14 @@ def _repr_sockets(sockets, mode):
     ret = []
     for i in sockets:
         url = ''
-        if isinstance(i, netlink_socket):
+        if isinstance(i, NetlinkSocket):
             ret.append('netlink://%s' % (i.family))
             continue
 
         if i.family == socket.AF_UNIX:
             url = 'unix'
+        elif i.family == AF_PIPE:
+            url = 'pipe'
         if type(i) == ssl.SSLSocket:
             if url:
                 url += '+'
@@ -173,6 +190,8 @@ def _repr_sockets(sockets, mode):
             url = 'tcp'
         if i.family == socket.AF_UNIX:
             url += '://%s' % (i.getsockname())
+        elif i.family == AF_PIPE:
+            url += '://%i,%i' % (i.getsockname())
         else:
             if mode == 'local':
                 url += '://%s:%i' % (i.getsockname())
@@ -182,7 +201,7 @@ def _repr_sockets(sockets, mode):
     return ret
 
 
-class marshal(object):
+class Marshal(object):
     '''
     Generic marshalling class
     '''
@@ -219,10 +238,10 @@ class marshal(object):
                 (length, msg_type) = struct.unpack('IH', self.buf.read(6))
                 error = None
                 if msg_type == NLMSG_ERROR:
-                    self.buf.seek(16)
+                    self.buf.seek(offset + 16)
                     code = abs(struct.unpack('i', self.buf.read(4))[0])
                     if code > 0:
-                        error = NetlinkSocketError(code)
+                        error = NetlinkError(code)
 
                 self.buf.seek(offset)
                 msg_class = self.msg_map.get(msg_type, nlmsg)
@@ -247,7 +266,43 @@ class marshal(object):
         pass
 
 
-class netlink_socket(socket.socket):
+class PipeSocket(object):
+    '''
+    Socket-like object for one-system IPC.
+
+    It is netlink-specific, since relies on length value
+    provided in the first four bytes of each message.
+    '''
+
+    family = AF_PIPE
+
+    def __init__(self, rfd, wfd):
+        self.rfd = rfd
+        self.wfd = wfd
+
+    def send(self, data):
+        os.write(self.wfd, data)
+
+    def recv(self, buf):
+        ret = os.read(self.rfd, 4)
+        length = struct.unpack('I', ret)[0]
+        ret += os.read(self.rfd, length - 4)
+        return ret
+
+    def getsockname(self):
+        return self.rfd, self.wfd
+
+    def fileno(self):
+        return self.rfd
+
+
+def pairPipeSockets():
+    pipe0_r, pipe0_w = os.pipe()
+    pipe1_r, pipe1_w = os.pipe()
+    return PipeSocket(pipe0_r, pipe1_w), PipeSocket(pipe1_r, pipe0_w)
+
+
+class NetlinkSocket(socket.socket):
     '''
     Generic netlink socket
     '''
@@ -282,12 +337,13 @@ class masq_record(object):
                                    self.pid, self.seq, time.ctime(self.ctime))
 
 
-class iothread(threading.Thread):
+class IOThread(threading.Thread):
     def __init__(self):
         threading.Thread.__init__(self)
         self.setDaemon(True)
         self.pid = os.getpid()
         self._nonce = 0
+        self._nonce_lock = threading.Lock()
         self._stop = False
         self._run_event = threading.Event()
         self._sctl_event = threading.Event()
@@ -307,18 +363,16 @@ class iothread(threading.Thread):
         self.uplinks = set()      # set(socket, socket...)
         self.servers = set()      # set(socket, socket...)
         self.controls = set()     # set(socket, socket...)
+        self.broadcast = set()    # set(socket, socket...)
         self.ssl_keys = {}        # {url: ssl_credentials(), url:...}
         self.mirror = False
         # secret; write non-zero byte as terminator
         self.secret = os.urandom(15)
         self.secret += '\xff'
-        # control socket, for in-process communication only
         self.uuid = uuid.uuid4()
-        self.sctl = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-        self.sctl.bind(b'\0%s' % (self.uuid))
-        self.sctl.listen(1)
-        self.servers.add(self.sctl)
-        self._rlist.add(self.sctl)
+        # control in-process communication only
+        self.sctl, self.control = pairPipeSockets()
+        self.add_client(self.sctl)
         self._sctl_thread = threading.Thread(target=self._sctl)
         self._sctl_thread.start()
         # masquerade cache expiration
@@ -335,8 +389,6 @@ class iothread(threading.Thread):
         self.backlog = []
 
     def _sctl(self):
-        self.control = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-        self.control.connect(b'\0%s' % (self.uuid))
         msg = ctrlmsg()
         msg['header']['type'] = NETLINK_UNUSED
         msg['cmd'] = IPRCMD_REGISTER
@@ -403,11 +455,12 @@ class iothread(threading.Thread):
             time.sleep(60)
 
     def nonce(self):
-        if self._nonce == 0xffffffff:
-            self._nonce = 1
-        else:
-            self._nonce += 1
-        return self._nonce
+        with self._nonce_lock:
+            if self._nonce == 0xffffffff:
+                self._nonce = 1
+            else:
+                self._nonce += 1
+            return self._nonce
 
     def distribute(self, data):
         """
@@ -439,7 +492,7 @@ class iothread(threading.Thread):
             target.socket.send(data.getvalue())
         else:
             # otherwise, broadcast packet
-            for sock in self.clients:
+            for sock in self.broadcast:
                 sock.send(data.getvalue())
             # return True -- this packet should be parsed
             return True
@@ -489,6 +542,7 @@ class iothread(threading.Thread):
                 if family in self.families:
                     send = self.families[family]
                     self.rtable[sock] = send
+                    self.broadcast.add(sock)
                     rsp['cmd'] = IPRCMD_ACK
                 # TODO tags: remote
                 #
@@ -535,12 +589,12 @@ class iothread(threading.Thread):
         Parse and enqueue messages. A message can be
         retrieved from netlink socket as well as from a
         remote system, and it should be properly enqueued
-        to make it available for netlink.get() method.
+        to make it available for Netlink.get() method.
 
-        If iothread.mirror is set, all messages will be also
+        If IOThread.mirror is set, all messages will be also
         copied (mirrored) to the default 0 queue. Please
         make sure that 0 queue exists, before setting
-        iothread.mirror to True.
+        IOThread.mirror to True.
 
         If there is no such queue for received
         sequence_number, leave sequence_number intact, but
@@ -554,10 +608,21 @@ class iothread(threading.Thread):
             msg['header']['host'] = _repr_sockets([sock], 'remote')[0]
             if key not in self.listeners:
                 key = 0
-            if self.mirror and key != 0:
-                self.listeners[0].put(copy.deepcopy(msg))
+            if self.mirror and (key != 0) and (msg.raw is not None):
+                # On Python 2.6 it can fail due to class fabrics
+                # in nlmsg definitions, so parse it again. It should
+                # not be much slower than copy.deepcopy()
+                try:
+                    self.listeners[0].put_nowait(marshal.parse(msg.raw)[0])
+                except Queue.Full:
+                    # FIXME: log this
+                    pass
             if key in self.listeners:
-                self.listeners[key].put(msg)
+                try:
+                    self.listeners[key].put_nowait(msg)
+                except Queue.Full:
+                    # FIXME: log this
+                    pass
 
     def command(self, cmd, attrs=[]):
         msg = ctrlmsg(io.BytesIO())
@@ -576,20 +641,22 @@ class iothread(threading.Thread):
     def reload(self):
         '''
         Reload I/O cycle. Warning: this method should be never
-        called from iothread, as it will not return.
+        called from IOThread, as it will not return.
         '''
         self._reload_event.clear()
         ret = self.command(IPRCMD_RELOAD)
         # wait max 3 seconds for reload
         # FIXME: timeout should be configurable
-        assert self._reload_event.wait(3)
+        self._reload_event.wait(3)
+        assert self._reload_event.is_set()
         return ret
 
     def add_server(self, url):
         '''
         Add a server socket to listen for clients on
         '''
-        (sock, address) = _get_socket(url, server_side=False)
+        (sock, address) = _get_socket(url, server_side=True,
+                                      ssl_keys=self.ssl_keys)
         if sock.family == socket.AF_INET:
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
         sock.bind(address)
@@ -641,11 +708,14 @@ class iothread(threading.Thread):
         self._rlist.remove(sock)
         self._wlist.remove(sock)
         self.clients.remove(sock)
+        if sock in self.broadcast:
+            self.broadcast.remove(sock)
         return sock
 
     def start(self):
         threading.Thread.start(self)
-        if not self._sctl_event.wait(3):
+        self._sctl_event.wait(3)
+        if not self._sctl_event.is_set():
             self._stop = True
             raise Exception('failed to establish control connection')
 
@@ -707,7 +777,7 @@ class iothread(threading.Thread):
                 self.route(fd, data)
 
 
-class netlink(object):
+class Netlink(object):
     '''
     Main netlink messaging class. It automatically spawns threads
     to monitor network and netlink I/O, creates and destroys message
@@ -716,28 +786,28 @@ class netlink(object):
     By default, netlink class connects to the local netlink socket
     on startup. If you prefer to connect to another host, use::
 
-        nl = netlink(host='tcp://remote.01host:7000')
+        nl = Netlink(host='tcp://remote.01host:7000')
 
     It is possible to connect to uplinks after the startup::
 
-        nl = netlink(do_connect=False)
+        nl = Netlink(do_connect=False)
         nl.connect('tcp://remote.01host:7000')
 
     To act as a server, call serve()::
 
-        nl = netlink(do_connect=False)
+        nl = Netlink(do_connect=False)
         nl.connect('localsystem')
         nl.serve('unix:///tmp/pyroute')
     '''
 
     family = NETLINK_GENERIC
     groups = 0
-    marshal = marshal
+    marshal = Marshal
 
     def __init__(self, debug=False, timeout=3, do_connect=True,
                  host='localsystem', key=None, cert=None, ca=None):
         self._timeout = timeout
-        self.iothread = iothread()
+        self.iothread = IOThread()
         self.listeners = self.iothread.listeners
         self.ssl_keys = self.iothread.ssl_keys
         self._sockets = set()
@@ -758,22 +828,43 @@ class netlink(object):
         smsg.encode()
         sock.send(smsg.buf.getvalue())
 
-    def connect(self, host='localsystem', key=None, cert=None, ca=None):
+    def connect(self,
+                host='localsystem',
+                key=None,
+                cert=None,
+                ca=None,
+                **kwarg):
+        assert isinstance(host, basestring)
         sock = None
         if key:
             self.ssl_keys[host] = ssl_credentials(key, cert, ca)
         try:
             if host == 'localsystem':
-                sock = netlink_socket(self.family)
+                sock = NetlinkSocket(self.family)
                 sock.bind(self.groups)
             else:
-                (sock, addr) = _get_socket(url=host,
-                                           server_side=False,
-                                           ssl_keys=self.ssl_keys)
-                sock.connect(addr)
+                try:
+                    # built-in connection types
+                    (sock, addr) = _get_socket(url=host,
+                                               server_side=False,
+                                               ssl_keys=self.ssl_keys)
+                    sock.connect(addr)
+                except AssertionError:
+                    # try to load a plugin
+                    (ctype, create, command) = _get_plugin(host)
+                    if ctype == 'queue':
+                        nonce = self.iothread.nonce()
+                        self.listeners[nonce] = create(command,
+                                                       self.marshal,
+                                                       **kwarg)
+                        return nonce
+                    else:
+                        raise Exception('plugin not supported')
+
                 self._remote_cmd(sock=sock,
                                  cmd=IPRCMD_ROUTE,
-                                 attrs=[['CTRL_ATTR_FAMILY_ID', self.family]])
+                                 attrs=[['CTRL_ATTR_FAMILY_ID',
+                                         self.family]])
                 rsp = ctrlmsg(sock.recv(28))
                 rsp.decode()
                 assert rsp['cmd'] == IPRCMD_ACK
@@ -796,8 +887,8 @@ class netlink(object):
 
     def shutdown_servers(self, *urls):
         self._shutdown_sockets([i for i in self.iothread.servers
-                                if _repr_sockets([i], 'remote') !=
-                                _repr_sockets([self.iothread.sctl], 'remote')],
+                                if _repr_sockets([i], 'local') !=
+                                _repr_sockets([self.iothread.sctl], 'local')],
                                'local', self.iothread.remove_server, *urls)
 
     def shutdown_clients(self, *urls):
@@ -821,6 +912,16 @@ class netlink(object):
                     pass
 
     def shutdown(self):
+        '''
+        Deprecated: use Netlink.release() instead
+        '''
+        logging.warn('IPDB: using deprecated call "shutdown()"')
+        self.release()
+
+    def release(self):
+        '''
+        Shutdown all threads and release netlink sockets
+        '''
         self.shutdown_sockets()
         self.shutdown_clients()
         self.shutdown_servers()
@@ -847,11 +948,11 @@ class netlink(object):
         messages that are not replies. They are just
         generated by the kernel as a reflection of settings
         changes. To start receiving these messages, call
-        netlink.monitor(). They can be fetched by
-        netlink.get(0) or just netlink.get().
+        Netlink.monitor(). They can be fetched by
+        Netlink.get(0) or just Netlink.get().
         '''
         if operate:
-            self.listeners[0] = Queue.Queue()
+            self.listeners[0] = Queue.Queue(maxsize=_QUEUE_MAXSIZE)
         else:
             del self.listeners[0]
 
@@ -871,7 +972,7 @@ class netlink(object):
                 if 0 in self.listeners:
                     self.listeners[0].put(msg)
 
-    def get(self, key=0):
+    def get(self, key=0, raw=False):
         '''
         Get a message from a queue
 
@@ -885,20 +986,27 @@ class netlink(object):
             # Bug-Url: http://bugs.python.org/issue1360
             try:
                 msg = queue.get(block=True, timeout=self._timeout)
-            except Queue.Empty:
+            except Queue.Empty as e:
+                if key == 0 or hasattr(queue, 'persist'):
+                    continue
                 self._remove_queue(key)
-                raise NetlinkQueueEmpty(key)
-            if msg['header']['error'] is not None:
+                raise e
+            # terminator for persisten queues
+            if msg is None:
+                self._remove_queue(key)
+                raise Queue.Empty()
+            if (msg['header']['error'] is not None) and (not raw):
                 self._remove_queue(key)
                 raise msg['header']['error']
-            if msg['header']['type'] != NLMSG_DONE:
+            if (msg['header']['type'] != NLMSG_DONE) or raw:
                 result.append(msg)
             if (msg['header']['type'] == NLMSG_DONE) or \
                (not msg['header']['flags'] & NLM_F_MULTI):
                 hosts -= 1
-            if hosts == 0:
+            if hosts == 0 or raw:
                 break
-        self._remove_queue(key)
+        if not hasattr(queue, 'persist'):
+            self._remove_queue(key)
         return result
 
     def send(self, buf):
@@ -908,7 +1016,7 @@ class netlink(object):
         '''
         data = buf.getvalue()
         for sock in self._sockets:
-            if isinstance(sock, netlink_socket):
+            if isinstance(sock, NetlinkSocket):
                 sock.sendto(data, (0, 0))
             else:
                 sock.send(data)
@@ -921,7 +1029,7 @@ class netlink(object):
         '''
         # FIXME make it thread safe, yeah
         nonce = self.iothread.nonce()
-        self.listeners[nonce] = Queue.Queue()
+        self.listeners[nonce] = Queue.Queue(maxsize=_QUEUE_MAXSIZE)
         msg['header']['sequence_number'] = nonce
         msg['header']['pid'] = os.getpid()
         msg['header']['type'] = msg_type
